#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/hw_random.h>
#include <asm/uaccess.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("NetApp, Inc");
MODULE_DESCRIPTION("RDRAND hwrng driver");

#define DEVICE_NAME "rdrand-rng"

/* 4.2.1 It is recommended that applications attempt 10 retries in a tight loop in the unlikely
   event that the RDRAND instruction does not return a random number. */
#define RDRAND_RETRY_LIMIT  10

int rdrand_read(struct hwrng* rng, void* data, size_t max, bool wait);
bool rdrand_supported(void);
inline bool rdrand64_step(u64* rand);
inline bool rdrand64_retry(u64* rand, unsigned int retries);

/* Hardware RNG operations */
static struct hwrng rdrand_rng =
{
    .name = DEVICE_NAME,
    .read = rdrand_read
};

/* Module init handler */
static int __init rdrand_init(void)
{
    if (!rdrand_supported())
        return -ENODEV;

    return hwrng_register(&rdrand_rng);
}

/* Module exit handler */
static void __exit rdrand_exit(void)
{
    hwrng_unregister(&rdrand_rng);
}

/* Hardware RNG read handler */
int rdrand_read(struct hwrng* rng, void* data, size_t max, bool wait)
{
    u64* data64 = data;
    size_t bytes_copied = 0;

    /* Read up to 'max' bytes (in 64-bit chunks), 'max' is guaranteed to be a multiple of 4
       and >= 32 bytes */
    while (bytes_copied + sizeof(u64) <= max)
    {
        /* Return failure regardless of 'wait' if the internal retry fails */
        if (!rdrand64_retry(data64++, RDRAND_RETRY_LIMIT))
            return -EIO;

        bytes_copied += sizeof(u64);
    }

    return bytes_copied;
}

/* Determine if CPU has RDRAND support */
bool rdrand_supported(void)
{
    unsigned int eax, ebx, ecx, edx;

    if (boot_cpu_data.x86_vendor != X86_VENDOR_INTEL)
        return false;

    /* Check for RDRAND support */
    cpuid(1, &eax, &ebx, &ecx, &edx);
    return (ecx & 0x40000000) == 0x40000000;
}

/* 64-bit invocation of RDRAND */
inline bool rdrand64_step(u64* rand)
{
    unsigned char ok = 0;
    asm volatile("rdrand %0; setc %1" : "=r"(*rand), "=qm"(ok));
    return ok;
}

/* 64-bit invocation of RDRAND with retry loop */
inline bool rdrand64_retry(u64* rand, unsigned int retries)
{
    unsigned int count = 0;
    while (count++ <= retries)
    {
        if (rdrand64_step(rand))
            return true;
    }

    return false;
}

module_init(rdrand_init);
module_exit(rdrand_exit);
